/*
	Class_DataList.cpp

	12/2/95
	As I noted in the header file, this is a relatively newly-written class.
	I have not had much time to thoroughly sieve for bugs, so be
	careful!
*/
#include "Class_DataList.h"
#include "assert_mac.h"

template<class NumEntriesType, class EntryLengthType>
DataList<NumEntriesType, EntryLengthType>::DataList() {
	this->fCurrentEntryIndex = 0;
	this->fCurrentEntryAddress = NULL;

	this->fData = (EntryHdl)NewHandle(sizeof(NumEntriesType));
	HLock((Handle)this->fData);
	**(NumEntriesType**)(this->fData) = 0;
} // END default ctor

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
DataList<NumEntriesType, EntryLengthType>::~DataList() {
	if (this->fData != NULL) {
		HUnlock((Handle)this->fData);
		DisposeHandle((Handle)this->fData);
		this->fData = NULL;
	}
} // END default dtor

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::SetData(EntryHdl data) {
	ASSERT(data != NULL);
	
	if (this->fData != NULL) {
		HUnlock((Handle)this->fData);
		DisposeHandle((Handle)this->fData);
	}

	this->fData = (EntryHdl)NewHandle(GetHandleSize((Handle)data));
	ASSERT(this->fData != NULL);
	HNoPurge((Handle)this->fData);
	MoveHHi((Handle)this->fData);
	HLock((Handle)this->fData);

	// Copy data
	BlockMove(*data, *this->fData, GetHandleSize((Handle)data));

	// Point to first entry
	this->fCurrentEntryIndex = 1;
	this->fCurrentEntryAddress = *(this->fData) + sizeof(NumEntriesType);
} // END SetData

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
EntryHdl DataList<NumEntriesType, EntryLengthType>::GetData() {
	return(this->fData);
} // END GetData

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
NumEntriesType DataList<NumEntriesType, EntryLengthType>::GetNumberOfEntries() {
	NumEntriesType *numEntriesPtr;
	
	numEntriesPtr = (NumEntriesType*)*this->fData;
	return(*numEntriesPtr);
} // END GetNumberOfEntries

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::SetNumberOfEntries(EntryIndex numEntries) {
	*(NumEntriesType*)*this->fData = numEntries;
} // END SetNumberOfEntries

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::SetCurrentEntryIndex(EntryIndex newIndex) {
	NumEntriesType numEntries = this->GetNumberOfEntries();
	
	ASSERT(newIndex > 0);
	ASSERT(newIndex <= numEntries);
	
	if (newIndex < 1 || newIndex > numEntries)
		return;
	
	this->fCurrentEntryIndex = newIndex;
	this->fCurrentEntryAddress = this->GetEntry(newIndex);
} // END SetCurrentEntryIndex

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
EntryPtr DataList<NumEntriesType, EntryLengthType>::GetCurrentEntry() {
	return(this->fCurrentEntryAddress);
} // END GetCurrentEntry

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
EntryPtr DataList<NumEntriesType, EntryLengthType>::GetNextEntry() {
	EntryPtr curEntry;
	NumEntriesType numEntries;
	
	numEntries = this->GetNumberOfEntries();
	curEntry = this->fCurrentEntryAddress;
	
	if (this->fCurrentEntryIndex < numEntries) {
		this->fCurrentEntryIndex++;
		
		// Move to next entry address
		EntryLengthType entryLength;
		entryLength = *(EntryLengthType*)this->fCurrentEntryAddress;
		this->fCurrentEntryAddress += sizeof(EntryLengthType) + entryLength;
	}
	else {
		this->fCurrentEntryIndex = 0;
		this->fCurrentEntryAddress = NULL;
	}

	return(curEntry);
} // END GetNextEntry

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
EntryPtr DataList<NumEntriesType, EntryLengthType>::GetEntry(EntryIndex index) {
	EntryPtr theEntry;
	NumEntriesType numEntries;
	EntryLengthType entryLength;
	EntryIndex i;
	
	numEntries = this->GetNumberOfEntries();

	ASSERT(index > 0);
	ASSERT(index <= numEntries);

	if (index < 1 || index > numEntries)
		return(NULL);
	
	// Parse through entry list
	theEntry = *(this->fData);
	// Move past number of entries
	theEntry += sizeof(NumEntriesType);
	
	for (i = 1; i < index; i++) {
		// Get current entry length
		entryLength = *(EntryLengthType*)theEntry;
		// Move past entry length and entry data
		theEntry += sizeof(EntryLengthType) + entryLength;
	}
	
	return(theEntry);
} // END GetEntry

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::SetEntry(EntryPtr entry, EntryIndex index) {
	EntryPtr theEntry;
	EntryLengthType oldEntrySize, newEntrySize;
	long oldSize, newSize;
	long offset;
	
	theEntry = this->GetEntry(index);
	oldEntrySize = *(EntryLengthType*)theEntry;
	newEntrySize = *(EntryLengthType*)entry;
	oldSize = GetHandleSize((Handle)this->fData);
	newSize = oldSize - oldEntrySize + newEntrySize;

	HUnlock((Handle)this->fData);
	SetHandleSize((Handle)this->fData, newSize);
	if (MemError() != noErr) {
	}
	HLock((Handle)this->fData);

	// Get theEntry address again, because it might
	// have changed due to SetHandleSize()
	theEntry = this->GetEntry(index);
	if (theEntry == NULL)
		return;
	
	offset = (char*)theEntry - (char*)*(this->fData);	// char* because we want bytes
		
	// Move old data forward/backward to make room for new data
	BlockMove(theEntry + oldEntrySize + sizeof(EntryLengthType),
		theEntry + newEntrySize + sizeof(EntryLengthType),
		oldSize - offset - oldEntrySize - sizeof(EntryLengthType));
	
	// Move new data in
	BlockMove(entry, theEntry, newEntrySize + sizeof(EntryLengthType));
	
	// Reset current entry address
	// (SetHandleSize() might have changed it)
	if (this->fCurrentEntryIndex != 0)
		this->SetCurrentEntryIndex(this->fCurrentEntryIndex);
	else
		this->SetCurrentEntryIndex(1);
} // END SetEntry

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::InsertEntry(EntryPtr entry, EntryIndex index) {
	EntryPtr theEntry;
	EntryLengthType oldEntrySize, newEntrySize;
	long oldSize, newSize;
	long offset;
	
	theEntry = this->GetEntry(index);
	oldEntrySize = *(EntryLengthType*)theEntry;
	newEntrySize = *(EntryLengthType*)entry;
	oldSize = GetHandleSize((Handle)this->fData);
	newSize = oldSize + newEntrySize + sizeof(EntryLengthType);

	HUnlock((Handle)this->fData);
	SetHandleSize((Handle)this->fData, newSize);
	if (MemError() != noErr) {
	}
	HLock((Handle)this->fData);

	theEntry = this->GetEntry(index);
	if (theEntry == NULL)
		return;
	
	offset = (char*)theEntry - (char*)*(this->fData);	// char* because we want bytes
	
	// Move old data forward/backward to make room for new data
	BlockMove(theEntry, theEntry + newEntrySize + sizeof(EntryLengthType),
		oldSize - offset + sizeof(EntryLengthType));
	
	// Move new data in
	BlockMove(entry, theEntry, newEntrySize + sizeof(EntryLengthType));
	
	// Update number of entries
	NumEntriesType numEntries;
	numEntries = this->GetNumberOfEntries();
	this->SetNumberOfEntries(numEntries+1);
	
	// Reset current entry address
	if (this->fCurrentEntryIndex != 0)
		this->SetCurrentEntryIndex(this->fCurrentEntryIndex);
	else
		this->SetCurrentEntryIndex(1);
} // END InsertEntry

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::AppendEntry(EntryPtr entry) {
	EntryPtr theEntry;
	EntryLengthType newEntrySize;
	NumEntriesType numEntries;
	long oldSize, newSize;
	
	newEntrySize = *(EntryLengthType*)entry;
	oldSize = GetHandleSize((Handle)this->fData);
	newSize = oldSize + newEntrySize + sizeof(EntryLengthType);

	HUnlock((Handle)this->fData);
	SetHandleSize((Handle)this->fData, newSize);
	if (MemError() != noErr) {
	}
	HLock((Handle)this->fData);
	
	// Update number of entries
	numEntries = this->GetNumberOfEntries();
	this->SetNumberOfEntries(numEntries+1);
	
	theEntry = (char*)*this->fData + oldSize;
	// Move new data in
	BlockMove(entry, theEntry, newEntrySize + sizeof(EntryLengthType));
	
	// Reset current entry address
	if (this->fCurrentEntryIndex != 0)
		this->SetCurrentEntryIndex(this->fCurrentEntryIndex);
	else
		this->SetCurrentEntryIndex(1);
} // END AppendEntry

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::DeleteEntry(EntryIndex index) {
	NumEntriesType numEntries;
	EntryPtr theEntry;
	EntryLengthType entrySize;
	long oldSize;
	long offset;
	
	numEntries = this->GetNumberOfEntries();
	ASSERT(numEntries > 0);
	if (numEntries < 1)
		return;

	oldSize = GetHandleSize((Handle)this->fData);
	theEntry = this->GetEntry(index);
	entrySize = *(EntryLengthType*)theEntry;
	offset = (char*)theEntry - (char*)*(this->fData);
	
	BlockMove(theEntry + entrySize + sizeof(EntryLengthType), theEntry,
		oldSize - offset - entrySize - sizeof(EntryLengthType));
	
	HUnlock((Handle)this->fData);
	SetHandleSize((Handle)this->fData, oldSize - entrySize - sizeof(EntryLengthType));
	if (MemError() != noErr) {
	}
	HLock((Handle)this->fData);

	// Update number of entries
	numEntries = this->GetNumberOfEntries();
	this->SetNumberOfEntries(numEntries - 1);

	// Reset current entry address
	// (SetHandleSize() might have changed it)
	if (this->GetNumberOfEntries() > 0)
		this->SetCurrentEntryIndex(1);
} // END DeleteEntry

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::SetEntry(DataPtr entryData, EntryLengthType length, EntryIndex index) {
	EntryPtr theEntry;
	
	theEntry = (EntryPtr)NewPtr(sizeof(EntryLengthType) + length);
	if (MemError() != noErr) {
	}
	
	*(EntryLengthType*)theEntry = length;
	BlockMove(entryData, (char*)theEntry + sizeof(EntryLengthType), length);
	
	this->SetEntry(theEntry, index);
	
	DisposePtr((Ptr)theEntry);
} // END SetEntry

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::InsertEntry(DataPtr entryData, EntryLengthType length, EntryIndex index) {
	EntryPtr theEntry;
	
	theEntry = (EntryPtr)NewPtr(sizeof(EntryLengthType) + length);
	if (MemError() != noErr) {
	}
	
	*(EntryLengthType*)theEntry = length;
	BlockMove(entryData, (char*)theEntry + sizeof(EntryLengthType), length);
	
	this->InsertEntry(theEntry, index);
	
	DisposePtr((Ptr)theEntry);
} // END InsertEntry

// ---------------------------------------------------------------------------

template<class NumEntriesType, class EntryLengthType>
void DataList<NumEntriesType, EntryLengthType>::AppendEntry(DataPtr entryData, EntryLengthType length) {
	EntryPtr theEntry;
	
	theEntry = (EntryPtr)NewPtr(sizeof(EntryLengthType) + length);
	if (MemError() != noErr) {
	}
	
	*(EntryLengthType*)theEntry = length;
	BlockMove(entryData, (char*)theEntry + sizeof(EntryLengthType), length);
	
	this->AppendEntry(theEntry);
	
	DisposePtr((Ptr)theEntry);
} // END AppendEntry
